@@ -1,4 +1,3 @@
-import sys
 import os
 import time
 import pytz
@@ -69,9 +68,14 @@ class EC2Manager:
             self._platform = self._config.get('platform', 'amzn2')
             self._architecture = self._config.get('architecture', 'arm64')
             self._private_subnet_cidr = self._config.get('private_subnet_cidr', '')
+            self._private_dns = self._config.get('private_dns', '')
+            self._ssh_key_name = self._config.get('ssh_key_name', '')
             self._instance_names = self._get_instance_names()
             self._instance_repo_updates = self._get_instance_repo_updates()
 
+            # holds all commands run on each instance in this session
+            self._command_execution_history = {instance_name: [] for instance_name in self._instance_names}
+
             # boost the pool size of urllib to handle as many concurrent connection as there are instances
             client_config = botocore.config.Config(
                 max_pool_connections=len(self._instance_names),
@@ -80,6 +84,7 @@ class EC2Manager:
             self._ssm_client = self._aws_session.client('ssm', config=client_config)
             self._ec2_client = self._aws_session.client('ec2', config=client_config)
             self._s3_client = self._aws_session.client('s3', config=client_config)
+            self._route53_client = self._aws_session.client('route53', config=client_config)
 
             self._github_client = Github(
                 login_or_token=self._github_token,
@@ -117,13 +122,9 @@ class EC2Manager:
         :rtype: dict
         """
         encoded_string = os.environ.get(encoded_env_data.get('env_var', ''), '')
-
-        # decode the env file
-        env_file_contents = base64.b64decode(encoded_string).decode('utf-8')
-        return {
-            line.split('=')[0]: line.split('=')[1].strip("'").strip('"') for line in env_file_contents.split('\n')
-            if '=' in line and not line.strip().startswith('#')
-        }
+        if encoded_string:
+            return json.loads(base64.b64decode(encoded_string).decode('utf-8'))
+        return {}
 
     @staticmethod
     def encode_string(string):
@@ -196,7 +197,7 @@ class EC2Manager:
         :return dict: A dictionary of instance names and their environment variables.
         """
         instances = self._config['instances']
-        default_ports = [{'from_port': 2375, 'to_port': 2375, 'protocol': 'tcp'}] if self._expose_docker_daemons else []
+        default_ports = [{'from_port': 22, 'to_port': 22, 'protocol': 'tcp'}] if self._ssh_key_name else []
 
         return {name: {
             'commands': {
@@ -204,6 +205,7 @@ class EC2Manager:
                 'stop': instances[name].get('commands', {}).get('stop', 'docker-compose down'),
                 'logs': instances[name].get('commands', {}).get('logs', 'docker-compose logs')
             },
+            'private_dns': f'{name}.{self._private_dns}' if self._private_dns else '',
             'volume_size': instances[name].get('volume_size', 8),
             'instance_type': instances[name].get('instance_type', 't4g.nano'),
             'working_directory': instances[name].get('working_directory', 'compose'),
@@ -278,9 +280,7 @@ class EC2Manager:
         if outputs.get('instances', {}).get('value') != self._instance_names:
             self._logger.info(f'Terraform instances...')
             instances = self.encode_string(json.dumps(self._get_instance_data()))
-
-            process = subprocess.run(
-                [
+            commands = [
                     'terraform',
                     'apply',
                     '-auto-approve',
@@ -300,7 +300,15 @@ class EC2Manager:
                     f'architecture={self._architecture}',
                     '-var',
                     f'instances={instances}'
-                ],
+                ]
+            if self._ssh_key_name:
+                commands.extend([
+                    '-var',
+                    f'ssh_key_name={self._ssh_key_name}'
+                ])
+
+            process = subprocess.run(
+                commands,
                 env=os.environ,
                 cwd=self._terraform_directory
             )
@@ -397,50 +405,9 @@ class EC2Manager:
             extra_args=['instance_status_ok']
         )
 
-    def _install_python(self, name):
-        if self._platform == 'amzn2':
-            self.run_command(
-                name,
-                ['sudo yum install python3 -y']
-            )
-        else:
-            raise NotImplemented
-
-    def _install_git(self, name):
-        if self._platform == 'amzn2':
-            self.run_command(
-                name,
-                ['sudo yum install git -y']
-            )
-        else:
-            raise NotImplemented
-
-    def _install_docker(self, name):
-        if self._platform == 'amzn2':
-            self.run_command(
-                name,
-                [
-                    'sudo amazon-linux-extras install docker',
-                    'sudo service docker start',
-                    'sudo usermod -a -G docker ssm-user',
-                    'sudo chkconfig docker on',
-                ])
-        else:
-            raise NotImplemented
-
-    def _install_docker_compose(self, name):
-        if self._platform == 'amzn2':
-            self.run_command(
-                name,
-                [
-                    'sudo curl -L https://github.com/docker/compose/releases/latest/download/docker-compose-$(uname -s)-$(uname -m) -o /usr/local/bin/docker-compose',
-                    'sudo chmod +x /usr/local/bin/docker-compose'
-                ])
-        else:
-            raise NotImplemented
-
     def _expose_docker_daemon(self, name):
         if self._expose_docker_daemons:
+            self._logger.info(f'Exposing docker daemon...')
             if self._platform == 'amzn2':
                 self.run_command(
                     name,
@@ -453,14 +420,6 @@ class EC2Manager:
             else:
                 raise NotImplemented
 
-    def _install_dependencies(self, name, data):
-        self._install_git(name)
-        self._install_python(name)
-        self._install_docker(name)
-        self._install_docker_compose(name)
-        self._expose_docker_daemon(name)
-
-
     def get_terraform_outputs(self):
         """
         Gets the output from terraform state.
@@ -502,7 +461,7 @@ class EC2Manager:
         :return dict: A dictionary of data for all instances in the region.
         """
         instances = {}
-        response = self._ec2_client.describe_instances()
+        response = self._ec2_client.describe_instances(MaxResults=500)
         for reservation in response["Reservations"]:
             for instance in reservation["Instances"]:
                 if instance['State']['Name'] == 'running':
@@ -527,6 +486,8 @@ class EC2Manager:
         """
         instance_id = self._get_instance_id(name, try_once=True)
         if instance_id:
+            # start from the home directory
+            commands.insert(0, f'cd /home')
             error = False
             # run the command
             response = self._ssm_client.send_command(
@@ -535,6 +496,7 @@ class EC2Manager:
                 Parameters={'commands': commands}
             )
             command_id = response['Command']['CommandId']
+            self._command_execution_history[name].extend(commands)
 
             # wait for the command to finish
             try:
@@ -581,7 +543,6 @@ class EC2Manager:
 
         # wait for each instance status to be ok
         self._await_instances()
-        self.install_dependencies()
 
     def _set_env_file(self, name, data):
         """
@@ -603,6 +564,41 @@ class EC2Manager:
                 f'echo "{env_file_content}" > .env'
             ], print_output=False)
 
+    def _set_dns(self, name, data):
+        """
+        Adds the ec2 ip to the dns.
+
+        :param str name: The name of the instance.
+        :param dict data: The instance's data.
+        """
+        private_dns = data['private_dns']
+        if private_dns:
+            private_ip_address = self.list_instances().get(name, {}).get('private_ip_address')
+            for hosted_zone in self._route53_client.list_hosted_zones().get('HostedZones', []):
+                if hosted_zone['Name'].strip('.') == private_dns.replace(f'{name}.', ''):
+                    self._logger.info(f'Setting {private_dns} to {private_ip_address}...')
+                    self._route53_client.change_resource_record_sets(
+                        HostedZoneId=hosted_zone['Id'],
+                        ChangeBatch={
+                            'Changes': [
+                                {
+                                    'Action': 'UPSERT',
+                                    'ResourceRecordSet': {
+                                        'Name': private_dns,
+                                        'Type': 'A',
+                                        'TTL': 300,
+                                        'ResourceRecords': [
+                                            {
+                                                'Value': private_ip_address
+                                            }
+                                        ]
+                                    }
+                                }
+                            ]
+                        }
+                    )
+                    break
+
     def _update_repo(self, name):
         """
         Updates a single instance repo.
@@ -612,11 +608,30 @@ class EC2Manager:
             self.run_command(name, [f'git clone {self._repo_url}'])
 
         self._logger.info(f'{name} checking out {self._commit}')
-        self.run_command(name, [
+        commands = [
             f'cd ./{self.repo}',
             'git fetch',
             'git pull'
-        ])
+        ]
+
+        # set the repo folder permissions for the default user
+        self._logger.info(f'{name} updating ./{self.repo} file permissions...')
+        if self._platform == 'ubuntu':
+            commands.insert(0, f'sudo setfacl -R -m u:ubuntu:rwx ./{self.repo}')
+        elif self._platform == 'amzn2':
+            commands.insert(0, f'sudo setfacl -R -m u:ec2-user:rwx ./{self.repo}')
+
+        self.run_command(name, commands)
+
+    def _update_user_data(self, name):
+        instance_id = self.list_instances().get(name, {}).get('instance_id')
+        response = self._ec2_client.describe_instance_attribute(
+            Attribute='userData',
+            DryRun=False,
+            InstanceId=instance_id
+        )
+        from pprint import pprint
+        pprint(response)
 
     def _stop(self, name, data):
         """
@@ -629,7 +644,7 @@ class EC2Manager:
             self._logger.info(f'{name} stopping...')
             working_directory = self._get_working_directory(data['working_directory'])
             self.run_command(name, [
-                f'cd {working_directory} ',
+                f'cd {working_directory}',
                 data['commands']['stop']
             ])
 
@@ -637,6 +652,7 @@ class EC2Manager:
         """
         Runs start command on a single instance.
         """
+        self._set_dns(name, data)
         self._set_env_file(name, data)
         self._logger.info(f'{name} starting...')
         working_directory = self._get_working_directory(data['working_directory'])
@@ -645,22 +661,13 @@ class EC2Manager:
             data['commands']['start'],
         ])
 
-    def install_dependencies(self):
-        """
-        Installs dependencies on the instances.
-        """
-        self._call_on_all(
-            callable_instance=self._install_dependencies,
-            items=self._get_instance_data(),
-        )
-
     def update_repos(self):
         """
         Updates the code in the instance repos.
         """
         self._call_on_all(
             callable_instance=self._update_repo,
-            items=self._instance_repo_updates,
+            items=self._get_instance_names(),
         )
 
     def stop(self):
